#include <iostream>
#include <fstream>
#include <sstream>
#include <list>
#include <map>
#include <set>
#include <string>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>


struct Research;
typedef std::map<std::string, Research*> ResearchMap;
typedef std::list<std::string> StringList;

static bool s_verbose = false;
static bool s_upedia = false;

struct Research
{
    Research(const std::string& key) 
        : m_key(key), m_cost(0), m_item(false), m_have(0), m_known(false),
          m_points(0), m_realcost(0), m_researched(0), m_brainers(0) { } 

    void addBase(const std::string& reqs);    
    void print(int level, bool complete, StringList& stack, ResearchMap& pmap, bool show_known);
    void freelist(bool show_known);
    void fulltree(ResearchMap& pmap);
    int  available(int level, bool noitem = false);

    std::string m_key;
    std::string m_name;
    std::string m_pedia;
    std::string m_fstr;

    int m_cost;
    bool m_item;
    int m_have;
    bool m_known;
    int m_points;
    
    int m_realcost;
    int m_researched;
    int m_brainers;

    ResearchMap m_unlocks;    // techs that unlock this one
    ResearchMap m_opens;      // techs unlocked by this
    ResearchMap m_depends;    // techs needed for this one
    ResearchMap m_needed;     // techs depending on this one
    ResearchMap m_getfree;    // researches that may grant this one.
    ResearchMap m_givefree;   // Can give free
};

struct NameSet
{
    std::string m_name;
    std::string m_pedia;
};

typedef std::map<std::string, NameSet*> NameMap;

struct RsrSet
{
    Research* addItem(const std::string& key);
    int parseResearch(const std::string& fileName);
    int translateItems(const std::string& fileName);
    int markKnownItems(const std::string& fileName);
    void printTranslated(const char* title, const std::string& key, bool pedia);
    void printMissions(bool pedia);

    int m_time[6];
    ResearchMap m_research;
    NameMap m_names;
};

void Research::addBase(const std::string& reqs)
{
    size_t pos = 0;
    while(pos != std::string::npos)
    {
        pos = reqs.find_first_not_of("[], ", pos);
        if(pos == std::string::npos) return;
        size_t endpos = reqs.find_first_of("[], ", pos);
        if(endpos == std::string::npos) return;
        m_fstr += ", ";
        m_fstr += reqs.substr(pos, endpos - pos);
        pos = endpos;
    }
}

void Research::print(int level, bool complete, StringList& stack, ResearchMap& pmap, bool show_known)
{
    // We want to prevent loops.
    // If this has been seen in this branch before, stop here.
    bool looped = false;

    if(level > 0 && m_known && !show_known) return;

    StringList::iterator st = stack.begin();
    while(st != stack.end())
    {
        if(*st == m_key)
        {
            looped = true;
            break;
        }
        ++st;
    }

    std::string offset(level * 3, ' ');
    std::string owned;
    if(m_have > 0)
    {
        std::stringstream ss;
        ss << "[owned: " << m_have << "] ";
        owned = ss.str();
    }
    std::cout << offset << m_name << " / " << m_key << " (" << m_cost << m_fstr << ") " << owned;
    int availability = 0;

    if(looped)
    {
        std::cout << " [ LOOPED ]" << std::endl;
        return;
    }
    else if(m_known)
    {
        std::cout << " [ KNOWN ]" << std::endl;
        return;
    }
    else if(pmap.find(m_key) != pmap.end())
    {
        std::cout  << " [ ABOVE ]" << std::endl;
        return;
    }
    else if(availability = available(1, true))
    {
        if(availability == 1) return;
    }
    if(availability != 2)  std::cout << std::endl;    
    if(!complete) return;

    stack.push_back(m_key);
    pmap[m_key] = this;

    ResearchMap::iterator it;
    bool nodes = false;
    
    if(!m_unlocks.empty() && availability == 0)
    {
        std::cout << offset << " UNLOCKED BY : " << std::endl;
        std::cout << offset << " {" << std::endl;
        for(it = m_unlocks.begin(); it != m_unlocks.end(); ++it)
        {
            it->second->print(level + 1, true, stack, pmap, show_known);
            nodes = true;
        }
        std::cout << offset << " }" << std::endl;
    }
    if(!m_depends.empty() && availability == 0)
    {
        std::cout << offset << " DEPENDS ON : " << std::endl;
        std::cout << offset << " {" << std::endl;
        for(it = m_depends.begin(); it != m_depends.end(); ++it)
        {
            it->second->print(level + 1, true, stack, pmap, show_known);
            nodes = true;
        }
        std::cout << offset << " }" << std::endl;
    }
    if(!m_getfree.empty())
    {
        std::cout << offset << " CAN BE GIVEN BY : " << std::endl;
        std::cout << offset << " {" << std::endl;
        for(it = m_getfree.begin(); it != m_getfree.end(); ++it)
        {
            it->second->print(level + 1, false, stack, pmap, true);
            nodes = true;
        }
        std::cout << offset << " }" << std::endl;
    }
    if(nodes) std::cout << offset << "/" << m_name << std::endl;
    stack.pop_back();
}

void Research::freelist(bool show_known)
{
    StringList stack;
    ResearchMap pmap;
    print(0, false, stack, pmap, show_known);
    ResearchMap::iterator it;
    if(!m_opens.empty())
    {
        std::cout << "UNLOCKS:" << std::endl;
        it = m_opens.begin();
        while(it != m_opens.end())
        {
            if(show_known || !it->second->m_known)
                it->second->print(1, false, stack, pmap, show_known);
            ++it;
        }
    }
    if(!m_needed.empty())
    {
        std::cout << "NEEDED FOR :" << std::endl;
        it = m_needed.begin();
        while(it != m_needed.end())
        {            
            if(show_known || !it->second->m_known)
            {
                it->second->print(1, false, stack, pmap, show_known);
                int n = it->second->m_depends.size() - 1;
                if(n > 0) std::cout << "    +" << n << " others" << std::endl;
            }
            ++it;
        }
    }

    if(!m_givefree.empty())
    {
        std::cout << "GIVES FREE :" << std::endl;
        it =  m_givefree.begin();
        while(it != m_givefree.end())
        {
            if(show_known || !it->second->m_known)
                it->second->print(1, false, stack, pmap, show_known);
            ++it;
        }
    }
}

void Research::fulltree(ResearchMap& pmap)
{
    // If self is in map, return.
    // Add self to map.
    // Scan unlocks, call fulltree for them
    // Scan depends, call fulltree for them
    ResearchMap::iterator it = pmap.find(m_key);
    if(it != pmap.end()) return;
    pmap[m_key] = this;

    it = m_opens.begin();
    while(it != m_opens.end())
    {
        if(!it->second->m_known) it->second->fulltree(pmap);
        ++it;
    }
    it = m_needed.begin();
    while(it != m_needed.end())
    {
        if(!it->second->m_known) it->second->fulltree(pmap);
        ++it;
    }
    it =  m_givefree.begin();
    while(it != m_givefree.end())
    {
        if(!it->second->m_known) it->second->fulltree(pmap);
        ++it;
    }
}

int Research::available(int level, bool noitem)
{
    // Already known 
    ResearchMap::iterator it;

    std::string header;
    if(level == 0)
    {
        std::stringstream ss;
        ss << m_name << " / " << m_key << " (" << m_cost << ") : ";
        header = ss.str();
    }    

    if(m_known && m_realcost == 0) 
    {
        // See if it has unknown give one free techs
        if(m_have > 0)
        {
            int giveFree = 0;
            it = m_givefree.begin();
            while(it != m_givefree.end())
            {
                if(!it->second->m_known) ++giveFree;
                ++it;
            }
            if(giveFree > 0) 
            {
                std::cout << header << "[ KNOWN but has " << giveFree << " free techs ]" << std::endl;
                return 2;
            }
        }
        return 1;
    }

    // Researching
    if(m_realcost > 0)
    {
        std::cout << header << "[ RESEARCHING ]"  << std::endl;
        return 1;
    }

    bool unlocked = false;
    // research needs to have at least one unlocked known
    if(!m_unlocks.empty())
    {
        it = m_unlocks.begin();
        while(it != m_unlocks.end())
        {
            if(it->second->m_known)
            {
                unlocked = true;
                break;
            }
            ++it;
        }
    }

    // or all depends must be known
    bool allDeps = true;
    if(!unlocked && !m_depends.empty())
    {
        it = m_depends.begin();
        while(it != m_depends.end())
        {
            if(!it->second->m_known) 
            {
                allDeps = false;
                break;
            }
            ++it;
        }
    }

    // Special case: has unlocks but no dependencies, must be unlocked.
    if(m_depends.empty() && !m_unlocks.empty())
    {
        if(!unlocked) allDeps = false;
    }

    if(allDeps)
    {
        // if noitem is true, list those that need but don't have item
        if(m_item)
        {
            if(m_have == 0)
            {
                if(noitem) 
                {
                    std::cout << header << "[ ITEM MISSING ]" << std::endl; 
                    return 2;
                }
                else return 2;
            }
            else
            {
                std::cout << header << "[ AVAILABLE, Have " << m_have << " ]" << std::endl; 
                return 2;
            }
        }
        else
        {
            std::cout << header << "[ AVAILABLE ]" << std::endl; 
            return 2;
        }
    }
    return 0;
}

bool hasValue(const std::string& line, const char* tag, std::string& val)
{
    size_t pos = line.find(tag);
    if(pos == std::string::npos) return false;
    pos += strlen(tag);
    pos = line.find_first_not_of(' ', pos);
    if(pos != std::string::npos) val = line.substr(pos);
    return true;
}

int itemQty(const std::string& line, std::string& val)
{
    size_t pos = line.find("STR_");
    if(pos == std::string::npos) return 0;
    size_t colpos = line.find(':', pos);
    if(colpos == std::string::npos) return 0;
    val = line.substr(pos, colpos - pos);
    return atoi(line.substr(colpos + 2).c_str());
}

Research* RsrSet::addItem(const std::string& key)
{
    ResearchMap::iterator it = m_research.find(key);
    if(it != m_research.end()) return it->second;
    Research* item = new Research(key);
    m_research[key] = item;
    return item;
}

void trimTrailing(std::string& line)
{
    size_t pos = line.find_last_not_of("\n\r\t ");
    if(pos != std::string::npos && pos < line.size())
    {
        line.erase(pos+1);
    }
}

int RsrSet::parseResearch(const std::string& fileName)
{
    int parsed = 0;
    std::ifstream ruleFile(fileName.c_str());

    if (!ruleFile)
    {
        std::cerr << "Unable to open " << fileName << std::endl;
        return parsed;
    }

    std::string sLine;
    while ( getline(ruleFile, sLine))
    {
        if(sLine.find("research:")  == 0) break;
    }

    Research* item = NULL;
    bool depends = false;
    bool unlocks = false;
    bool getfree = false;
    
    while ( getline(ruleFile, sLine))
    {
        trimTrailing(sLine);
        if(sLine[0] != ' ' && sLine[0] != '#') break; // Next section start
        std::string val;
        if(hasValue(sLine, "- name:", val)) // start new research
        {            
            ++parsed;         
            item = addItem(val);            
            depends = false;
            unlocks = false;
            getfree = false;
            continue;
        }
        if(item)
        {
            if(depends)
            {
                if(hasValue(sLine, "-", val))
                {
                    Research* dep = addItem(val);                    
                    item->m_depends[val] = dep;
                    dep->m_needed[item->m_key] = item;
                    continue; 
                }
                else depends = false;
            }
            else if(unlocks)
            {
                if(hasValue(sLine, "-", val))
                {
                    Research* dep = addItem(val);
                    dep->m_unlocks[item->m_key] = item;
                    item->m_opens[val] = dep;                    
                    continue;
                }
                else unlocks = false;
            }
            else if(getfree)
            {
                if(hasValue(sLine, "-", val))
                {
                    Research* dep = addItem(val);
                    dep->m_getfree[item->m_key] = item;
                    item->m_givefree[val] = dep;
                    continue;
                }
                else getfree = false;
            }
            // Fell through here
            if(hasValue(sLine, "dependencies:", val))
            {
                depends = true;                
            }
            else if(hasValue(sLine,"unlocks:", val))
            {
                unlocks = true;
            }
            else if(hasValue(sLine,"getOneFree:", val))
            {
                getfree = true;
            }
            else if(hasValue(sLine,"cost:", val))
            {
                item->m_cost = atoi(val.c_str());
            }
            else if(hasValue(sLine,"points:", val))
            {
                item->m_points = atoi(val.c_str());
            }
            else if(hasValue(sLine,"needItem:", val))
            {
                item->m_item = (val == "true");
            }
            else if(hasValue(sLine,"lookup:", val))
            {
                Research* dep = addItem(val);
                dep->m_unlocks[item->m_key] = item;
                item->m_opens[val] = dep;
            }
            else if(hasValue(sLine,"requiresBaseFunc:", val))
            {
                item->addBase(val);
            }
        }
    }

    return parsed;
}

int RsrSet::translateItems(const std::string& fileName)
{
    int parsed = 0;
    int rtrans = 0;
    int ntrans = 0;
    std::ifstream langFile(fileName.c_str());
    if (!langFile)
    {
        std::cerr << "Unable to open " << fileName << std::endl;
        return parsed;
    }

    std::string sLine;
    while ( getline(langFile, sLine))
    {
        trimTrailing(sLine);
        size_t pos = sLine.find("STR_");
        if(pos == std::string::npos) continue;
        size_t colpos = sLine.find(':', pos);
        if(colpos == std::string::npos) continue;
        size_t q1pos = sLine.find_first_not_of(": ", colpos);
        if(q1pos == std::string::npos) continue;
        size_t q2pos;
        if(sLine[q1pos] == '"') // Value is in quotes
        {
            ++q1pos;
            q2pos = sLine.find_first_of('"', q1pos);
            if(q2pos == std::string::npos) continue;
        }
        else // No quotes
        {
            q2pos = sLine.find_last_not_of(" \n\r\t");
            if(q2pos < q1pos) continue;
            q2pos++;
        }
        ++parsed;
        std::string key = sLine.substr(pos, colpos - pos);
        std::string val = sLine.substr(q1pos, q2pos - q1pos);

        // Is it an ufopedia entry?
        size_t upos = key.find("_UC_UFOPEDIA");      
        if(upos == std::string::npos) upos = key.find("_UFOPEDIA"); 
       
        if(upos != std::string::npos)
        {
            size_t nlpos = 0;
            while((nlpos = val.find("{NEWLINE}")) != std::string::npos)
            {
                if(nlpos == 0 || (nlpos > 0 && val[nlpos - 1] == '\n')) 
                    val.erase(nlpos, 9);
                else 
                    val.replace(nlpos, 9, "\n");
            } 
        }

        ResearchMap::iterator it ;
        if(upos != std::string::npos)
        {
            it = m_research.find(key.substr(0, upos));
        }
        else
        {
            it = m_research.find(key);
        }    

        if(it != m_research.end())
        {
            if(upos == std::string::npos)
            {
                ++rtrans;
                it->second->m_name = val;
            }
            else 
            {                
                it->second->m_pedia = val;
            }
        }

        NameMap::iterator nit;
        if(upos != std::string::npos)
        {
            nit = m_names.find(key.substr(0, upos));
        }
        else
        {
            nit = m_names.find(key);
        }    
        if(nit != m_names.end())
        {
            NameSet* nset = nit->second;
            if(!nset) 
            {
                nset = new NameSet();
                nit->second = nset;
                ++ntrans;
            }
            if(upos != std::string::npos)
                nset->m_pedia = val;
            else 
                nset->m_name = val;
        }
    }

    ResearchMap::iterator it = m_research.begin();
    while(it != m_research.end())
    {
        if(it->second->m_name.empty()) 
        {
            ++rtrans;
            if(it->first.substr(0,4) == "STR_")
                it->second->m_name = it->first.substr(4);
            else
                it->second->m_name = it->first;
        }
        ++it;
    }
    
    if(s_verbose) std::cout << "Research translated: " << rtrans << std::endl;

    return rtrans;    
}

int RsrSet::markKnownItems(const std::string& fileName)
{
    int known = 0;
    std::ifstream saveFile(fileName.c_str());
    if (!saveFile)
    {
        std::cerr << "Unable to open " << fileName << std::endl;
        return known;
    }

    std::string sLine;
    std::string val;
    
    bool haveTime = false;
    while( getline(saveFile, sLine) )
    {
        if(sLine.find("time:") == 0) haveTime = true;
        else if(sLine[0] != ' ' && haveTime) break;
        else if(haveTime)
        {
            if(hasValue(sLine,"second:", val)) m_time[5] = atoi(val.c_str());
            else if(hasValue(sLine,"minute:", val)) m_time[4] = atoi(val.c_str());
            else if(hasValue(sLine,"hour:", val)) m_time[3] = atoi(val.c_str());
            else if(hasValue(sLine,"day:", val)) m_time[2] = atoi(val.c_str());
            else if(hasValue(sLine,"month:", val)) m_time[1] = atoi(val.c_str());
            else if(hasValue(sLine,"year:", val)) m_time[0] = atoi(val.c_str());
        }
    }

    if(s_verbose)
    {
        char timestr[50];
        sprintf(timestr, "SaveFile Time : %04d-%02d-%02d %02d:%02d:%02d", 
                m_time[0], m_time[1], m_time[2], m_time[3], m_time[4], m_time[5]);
        std::cout << timestr << std::endl;
    }

    // Locate bases info
    while ( getline(saveFile, sLine))
    {
        if(sLine.find("bases:") == 0) break;
    }
    
    bool haveitems = false;
    bool haveproject = false;
    ResearchMap::iterator it;
    int appliedItems = 0;
    int totalPoints = 0;
    int free_brainers = 0;
    while ( getline(saveFile, sLine))
    {
        trimTrailing(sLine);
        if(sLine[0] != ' ') break;
        if(haveitems)
        {
            int qty = itemQty(sLine, val);
            if(qty != 0)
            {
                it = m_research.find(val);
                if(it != m_research.end()) 
                {            
                    if(it->second->m_have == 0) ++appliedItems;  
                    it->second->m_have += qty;
                }
                else
                {
                    // Create fake research item
                    Research* item = addItem(val);
                    item->m_have += qty;
                    item->m_known = true;
                }
                continue;
            }
            else 
            {
                haveitems = false;
            }
        }
        if(haveproject)
        {
            if(hasValue(sLine,"assigned:", val))
            {
                it->second->m_brainers = atoi(val.c_str());
            }
            else if(hasValue(sLine,"spent:", val))
            {
                it->second->m_researched = atoi(val.c_str());
            }
            else if(hasValue(sLine,"cost:", val))
            {
                it->second->m_realcost = atoi(val.c_str());
            }
            else haveproject = false;
        }
        if((sLine.find("items:") != std::string::npos))
        {
            haveitems = true;
        }
        else if(hasValue(sLine,"- project:", val))
        {
            it = m_research.find(val);
            if(it != m_research.end()) haveproject = true;
        }
        else if(hasValue(sLine,"scientists:", val))
        {
            free_brainers += atoi(val.c_str());
        }
    }

    if(s_verbose) std::cout << "Applied " << appliedItems << " inventory items" << std::endl;

    
    // find 'discovered:'
    while ( getline(saveFile, sLine))
    {
        if(sLine.find("discovered:") == 0) break;
    }

    while ( getline(saveFile, sLine))
    {
        trimTrailing(sLine);
        if(hasValue(sLine, "-", val))
        {         
            ResearchMap::iterator it = m_research.find(val);
            if(it != m_research.end()) 
            {
                it->second->m_known = true;
                totalPoints += it->second->m_points;
                ++known;
            }
        }
        else break;
    }

    if(s_verbose) std::cout << "Earned " << totalPoints << " points from research" << std::endl;

    if(free_brainers > 0) 
        std::cout << "WARNING: " << free_brainers << " Unassigned brainers!" << std::endl;

    return known;
}

void RsrSet::printTranslated(const char* title, const std::string& key, bool pedia)
{
    NameMap::iterator nit = m_names.find(key);
    if(nit != m_names.end() && nit->second && !nit->second->m_name.empty())
    {
        std::cout << title << nit->second->m_name << std::endl;
    }
    else
    {
        std::cout << title << key << std::endl;
    }
    if(pedia && nit != m_names.end() && nit->second && !nit->second->m_pedia.empty())
    {
        std::cout << "  > " << nit->second->m_pedia << std::endl;
    }
}

int main(int argc, char **argv)
    
{
    // Steps:
    // Open the rules file
    // Scan and parse all manufactures, building item map;
    // Scan and parse items, any item in item map, assign buy and sell prices.
    // Scan lang file, apply translations to items in map.
    // Analyze and calculate prices.
    RsrSet rsrSet;

    std::string saveDir("./user/piratez/");
    std::string ruleDir("./user/mods/Piratez/Ruleset/");
    
    std::string ruleFileName("Piratez.rul");
    std::string langFileName("Piratez_lang.rul");
    std::string saveFileName("game");
    
    std::string searchStr;
    int speccost = 0;
    char option = '\0';;
    bool isdep = false;
    bool iscost = false;
    bool isfc = false;
    bool isavail = false;
    bool isreg = true;
    bool noitem = false;
    bool pedia = false;
    bool current = false;
    bool show_known = false;
    bool fulltree = false;

    for(int i = 1; i < argc; ++i)
    {
        if(argv[i][0] == '-') // option
        {
            option = argv[i][1];
            if(option == 'g') isdep = true;
            if(option == 'f') isfc = true;  
            if(option == 'c') iscost = true;  
            if(option == 'a') isavail = true;  
            if(option == 'v') s_verbose = true;
            if(option == 'k') show_known = true;
            if(option == 'u') pedia = true;
            if(option == 'n') noitem = true;
            if(option == 'p') current = true;
            if(option == 't') fulltree = true;
            continue;
        }
        switch(option)
        {
        case 'r': // rule file name
            ruleFileName.assign(argv[i]);
            break;
        case 'l': // lang file name
            langFileName.assign(argv[i]);
            break;
        case 's': // save file name
            saveFileName.assign(argv[i]);
            break;
        case 'c': // cost analisys
            speccost = atoi(argv[i]);
            break;
        case 'a':
            if(argv[i][0] == 'N') noitem = true;
            break;
        default:
            searchStr.assign(argv[i]);     
        }
        option = '\0';
    }

    if(iscost || isdep || isavail || isfc || current || fulltree) isreg = false;
    if((isdep || isreg || fulltree) && searchStr.empty()) 
    {
        std::cout << "Error: no subject provided" << std::endl;
        exit(1);
    }
    
    if(s_verbose) 
    {
        std::cout << "Rule file : " <<  ruleFileName << std::endl;
        std::cout << "Lang file : " <<  langFileName << std::endl;
        std::cout << "Save file : " <<  saveFileName << std::endl;
    
        if(iscost)
        {
            std::cout << "Cost Analisys mode [" << speccost << "]" << std::endl;
        }
        else if(isfc)
        {
            std::cout << "Facility dependency mode [" << searchStr << "]" << std::endl;
        }
        else if(isavail)
        {
            std::string aux("Have Item");
            if(noitem) aux.assign("Any");
            std::cout << "Availability mode [" << aux << "]" << std::endl;
        }
        else if(isdep)
        {
            std::cout << "Dependency mode [" << searchStr << "]" << std::endl;
        }
        else if(fulltree)
        {
            std::cout << "Full dependency tree mode [" << searchStr << "]" << std::endl;
        }
        else
        {
            std::cout << "Regular check mode [" << searchStr << "]" << std::endl;
        }
    }

    std::string ruleFile(ruleDir);
    ruleFile += ruleFileName;
    int nparsed = rsrSet.parseResearch(ruleFile);
    if(s_verbose) std::cout << "Parsed "  << nparsed << " lines" << std::endl;

    std::string saveFile(saveDir);
    saveFile += saveFileName;
    saveFile += ".sav";
    int known = rsrSet.markKnownItems(saveFile);
    if(s_verbose) std::cout << "Marked "  << known << " as known" << std::endl;
    
    std::string langFile(ruleDir);
    langFile += langFileName;
    nparsed = rsrSet.translateItems(langFile);

    std::map<int, int> costMap;
    StringList stack;
    ResearchMap pmap; 
    int totalPoints = 0;
    int earnedPoints = 0;
    ResearchMap::iterator it;
    if(isreg || isdep || fulltree) 
    {   
        it = rsrSet.m_research.find(searchStr);   
    }
    else it = rsrSet.m_research.begin();
    
    if(it == rsrSet.m_research.end()) std::cout << "Subject " << searchStr << " not found." << std::endl;
    
    int ncounted = 0;
    while(it != rsrSet.m_research.end())
    {        
        if(iscost)
        {
            if(show_known || !it->second->m_known)
            {
                ++ncounted;
                std::map<int, int>::iterator cit = costMap.find(it->second->m_cost);
                if(cit != costMap.end()) ++(cit->second);
                else costMap[it->second->m_cost] = 1;
                if(speccost && it->second->m_cost == speccost)
                    it->second->print(1, false, stack, pmap, show_known);
            }
            totalPoints += it->second->m_points;
            if (it->second->m_known) earnedPoints += it->second->m_points;
        }
        else if(isfc) // no specific facility support yet
        {
            if(!it->second->m_fstr.empty())
                it->second->print(0, false, stack, pmap, show_known);
        }
        else if(current)
        {
            if(it->second->m_realcost > 0) it->second->available(0, noitem);
        }
        else if(isavail)
        {
            it->second->available(0, noitem);
        }
        else if(isdep)
        {
            it->second->freelist(show_known);
            break;
        }
        else if(fulltree)
        {
            it->second->print(0, false, stack, pmap, show_known);
            it->second->fulltree(pmap);
            ResearchMap::iterator tt = pmap.begin();
            ResearchMap dmap;
            while(tt != pmap.end())
            {
                if(!tt->second->m_known && tt->second != it->second) 
                    tt->second->print(0, false, stack, dmap, show_known);
                ++tt;
            }
            break;
        }
        else
        {
            it->second->print(0, true, stack, pmap, show_known);
            break;
        }
        ++it;
    }
    if(pedia && it != rsrSet.m_research.end()) 
        std::cout << "   " << it->second->m_pedia << std::endl;

    if(iscost)
    {
        std::cout << "Found " << ncounted << " unknown projects." << std::endl;
        int running = 0;
        int ndays = 0;
        std::map<int, int>::iterator cit = costMap.begin();
        while(cit != costMap.end())
        {
            if(cit->first == 0)
            {
                ncounted -= cit->second;
                std::cout << cit->second << " projects with 0 cost - not counted" << std::endl;
            }
            else
            {
                running += cit->second;
                ndays += cit->first * cit->second;
                double pct = (100.00 * running) / ncounted;
                std::cout << cit->first << " : " << cit->second << " (" << running << ") " << pct << "%" << std::endl;
            }
            ++cit;
        }
        std::cout << "Total: " << ndays << std::endl;
        std::cout << earnedPoints << " points earned out of " << totalPoints << std::endl;
    }

    return 0;
}
